// David Pilling 11th December 2015
// liquid level sensor with ultrasound



#include "LowPower.h"



#include <stdarg.h>

void pr(char *fmt, ... ){
        char buf[128]; // resulting string limited to 128 chars
        va_list args;
        va_start (args, fmt );
        vsnprintf(buf, 128, fmt, args);
        va_end (args);
        Serial.println(buf);

}




#define pin_led 13
#define pin_sensor_power 12
#define pin_power_off 11
#define SWITCH_PIN 11

#define pin_temp_power 10
#define pin_temp_signal 9
#define pin_tx_power 8
#define pin_tx_data (7)

 
#define TRIGGER_PIN 5
#define AIN0 6
#define AIN1 7
#define CHARGE_PIN 4









int mode_speed=1;  // bigger == slower
int mode_slowspeed=4;
int mode_count=0;
int mode_switch=100; // after mode_count equals this switch to slow mode
int mode_slow=0;     // true when in slow mode



/*
 * connectingStuff, Oregon Scientific v2.1 Emitter
 * http://connectingstuff.net/blog/encodage-protocoles-oregon-scientific-sur-arduino/
 *
 * Copyright (C) 2013 olivier.lebrun@gmail.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/


void writeBCD(byte * p,int nybble,int width,long value)
{
 char buffer[16]; 
 char fmt[8];
 int  byteindex=nybble/2;
 int  i;
 int  c;

 nybble=nybble-byteindex*2; 
  
  
// Serial.println(value); 

 sprintf(fmt,"%%0%dd",width); 
 sprintf(buffer,fmt,value);

// Serial.println(buffer);


 for(i=0;i<width;i++)
 {
  c=buffer[i]-'0';
//   Serial.println(c);
  if(nybble & 0x1)
  {
   p[byteindex]=(p[byteindex] & 0xF0) | c;
   nybble=0;
   byteindex++;
  }
  else
  {
   p[byteindex]=(p[byteindex] & 0xF) | (c<<4);
   nybble++;
  } 
 } 
}





const unsigned long TIME = 512;
const unsigned long TWOTIME = 1024;
 
#define SEND_HIGH() digitalWrite(pin_tx_data, 1)
#define SEND_LOW() digitalWrite(pin_tx_data, 0)



 
// Buffer for Oregon message
  byte OregonMessageBuffer[9];

 
/**
 * \brief    Send logical "0" over RF
 * \details  azero bit be represented by an off-to-on transition
 * \         of the RF signal at the middle of a clock period.
 * \         Remenber, the Oregon v2.1 protocol add an inverted bit first 
 */
inline void sendZero(void) 
{
  SEND_HIGH();
  delayMicroseconds(TIME);
//  pulse(TIME,HIGH);
  SEND_LOW();
  delayMicroseconds(TWOTIME);
//  pulse(TWOTIME,LOW);
  SEND_HIGH();
  delayMicroseconds(TIME);
//  pulse(TIME,HIGH);  
}
 
/**
 * \brief    Send logical "1" over RF
 * \details  a one bit be represented by an on-to-off transition
 * \         of the RF signal at the middle of a clock period.
 * \         Remenber, the Oregon v2.1 protocol add an inverted bit first 
 */
inline void sendOne(void) 
{
   SEND_LOW();
   delayMicroseconds(TIME);
//   pulse(TIME,LOW);   
   SEND_HIGH();
   delayMicroseconds(TWOTIME);
//   pulse(TWOTIME,HIGH);   
   SEND_LOW();
   delayMicroseconds(TIME);
//   pulse(TIME,LOW);   
}
 
/**
* Send a bits quarter (4 bits = MSB from 8 bits value) over RF
*
* @param data Source data to process and sent
*/
 
/**
 * \brief    Send a bits quarter (4 bits = MSB from 8 bits value) over RF
 * \param    data   Data to send
 */
inline void sendQuarterMSB(const byte data) 
{
  (bitRead(data, 4)) ? sendOne() : sendZero();
  (bitRead(data, 5)) ? sendOne() : sendZero();
  (bitRead(data, 6)) ? sendOne() : sendZero();
  (bitRead(data, 7)) ? sendOne() : sendZero();
}
 
/**
 * \brief    Send a bits quarter (4 bits = LSB from 8 bits value) over RF
 * \param    data   Data to send
 */
inline void sendQuarterLSB(const byte data) 
{
  (bitRead(data, 0)) ? sendOne() : sendZero();
  (bitRead(data, 1)) ? sendOne() : sendZero();
  (bitRead(data, 2)) ? sendOne() : sendZero();
  (bitRead(data, 3)) ? sendOne() : sendZero();
}
 
/******************************************************************/
/******************************************************************/
/******************************************************************/
 
/**
 * \brief    Send a buffer over RF
 * \param    data   Data to send
 * \param    size   size of data to send
 */
void sendData(byte *data, byte size)
{
  for(byte i = 0; i < size; ++i)
  {
    sendQuarterLSB(data[i]);
    sendQuarterMSB(data[i]);
  }
}
 
/**
 * \brief    Send an Oregon message
 * \param    data   The Oregon message
 */
void sendOregon(byte *data, byte size)
{
    sendPreamble();
    //sendSync();
    sendData(data, size);
    sendPostamble();
}
 
/**
 * \brief    Send preamble
 * \details  The preamble consists of 16 "1" bits
 */
inline void sendPreamble(void)
{
  byte PREAMBLE[]={0xFF,0xFF};
  sendData(PREAMBLE, 2);
}
 
/**
 * \brief    Send postamble
 * \details  The postamble consists of 8 "0" bits
 */
inline void sendPostamble(void)
{
  byte POSTAMBLE[]={0x00};
  sendData(POSTAMBLE, 1);  

}
 
/**
 * \brief    Send sync nibble
 * \details  The sync is 0xA. It is not use in this version since the sync nibble
 * \         is include in the Oregon message to send.
 */
inline void sendSync(void)
{
  sendQuarterLSB(0xA);
}
 
/******************************************************************/
/******************************************************************/
/******************************************************************/
 
/**
 * \brief    Set the sensor type
 * \param    data       Oregon message
 * \param    type       Sensor type
 */
inline void setType(byte *data, byte* type) 
{
  data[0] = type[0];
  data[1] = type[1];
}
 
/**
 * \brief    Set the sensor channel
 * \param    data       Oregon message
 * \param    channel    Sensor channel (0x10, 0x20, 0x30)
 */
inline void setChannel(byte *data, byte channel) 
{
    data[2] = channel;
}
 
/**
 * \brief    Set the sensor ID
 * \param    data       Oregon message
 * \param    ID         Sensor unique ID
 */
inline void setId(byte *data, byte ID) 
{
  data[3] = ID;
}
 
/**
 * \brief    Set the sensor battery level
 * \param    data       Oregon message
 * \param    level      Battery level (0 = low, 1 = high)
 */
void setBatteryLevel(byte *data, byte level)
{
  if(!level) data[4] = 0x0C;
  else data[4] = 0x00;
}
 
/**
 * \brief    Set the sensor temperature
 * \param    data       Oregon message
 * \param    temp       the temperature
 */
void setTemperature(byte *data, float temp) 
{
  // Set temperature sign
  if(temp < 0)
  {
    data[6] = 0x08;
    temp *= -1;  
  }
  else
  {
    data[6] = 0x00;
  }
 
  // Determine decimal and float part
  int tempInt = (int)temp;
  int td = (int)(tempInt / 10);
  int tf = (int)round((float)((float)tempInt/10 - (float)td) * 10);
 
  int tempFloat =  (int)round((float)(temp - (float)tempInt) * 10);
 
  // Set temperature decimal part
  data[5] = (td << 4);
  data[5] |= tf;
 
  // Set temperature float part
  data[4] |= (tempFloat << 4);
}
 
/**
 * \brief    Set the sensor humidity
 * \param    data       Oregon message
 * \param    hum        the humidity
 */
void setHumidity(byte* data, byte hum)
{
    data[7] = (hum/10);
    data[6] |= (hum - data[7]*10) << 4;
}
 
/**
 * \brief    Sum data for checksum
 * \param    count      number of bit to sum
 * \param    data       Oregon message
 */
int Sum(byte count, const byte* data)
{
  int s = 0;
 
  for(byte i = 0; i<count;i++)
  {
    s += (data[i]&0xF0) >> 4;
    s += (data[i]&0xF);
  }
 
  if(int(count) != count)
    s += (data[count]&0xF0) >> 4;
 
  return s;
}
 
/**
 * \brief    Calculate checksum
 * \param    data       Oregon message
 */
void calculateAndSetChecksum(byte* data)
{
 data[8] = ((Sum(8, data) - 0xa) & 0xFF);
}

void transmitter_setup(void)
{
  byte ID[] = {0x1A,0xBC};
  
  pinMode(pin_tx_data , OUTPUT);
  SEND_LOW();  
//  pulse(0,LOW);
 
  setType(OregonMessageBuffer, ID);
  setChannel(OregonMessageBuffer, 0x20);
  setId(OregonMessageBuffer, 0xBB);
  
}


void transmitter_send(void)
{
  calculateAndSetChecksum(OregonMessageBuffer);
 
 // Serial.println("send message");
  // Show the Oregon Message
  for (byte i = 0; i < sizeof(OregonMessageBuffer); ++i)   
  {
//    Serial.print(OregonMessageBuffer[i] >> 4, HEX);
//    Serial.print(OregonMessageBuffer[i] & 0x0F, HEX);
  }
 
  Serial.println();
  // Send the Message over RF
  sendOregon(OregonMessageBuffer, sizeof(OregonMessageBuffer));
  // Send a "pause"
  SEND_LOW();
//  pulse(0,LOW);

  delayMicroseconds(TWOTIME*8);
  // Send a copie of the first message. The v2.1 protocol send the
  // message two time 
  sendOregon(OregonMessageBuffer, sizeof(OregonMessageBuffer));
 
  // Wait for 30 seconds before send a new message 
  SEND_LOW();  
 
}




// https://code.google.com/p/tinkerit/wiki/SecretThermometer

long readTemp() {
  long result;
  // Read temperature sensor against 1.1V reference
  ADMUX = _BV(REFS1) | _BV(REFS0) | _BV(MUX3);
  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA,ADSC));
  result = ADCL;
  result |= ADCH<<8;
  result = (result - 125) * 1075;
  return result;
}


// https://code.google.com/p/tinkerit/wiki/SecretVoltmeter

long readVcc() {
  long result;
  // Read 1.1V reference against AVcc
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA,ADSC));
  result = ADCL;
  result |= ADCH<<8;
  result = 1126400L / result; // Back-calculate AVcc in mV

  ADMUX=0;

  return result;
}



#include <dht11.h>

void powerup_temp(void)
{
 pinMode(pin_temp_power, OUTPUT);  
 digitalWrite(pin_temp_power, HIGH);
}


void powerdown_temp(void)
{
 pinMode(pin_temp_signal, INPUT);   
 digitalWrite(pin_temp_signal, LOW);

 digitalWrite(pin_temp_power, LOW);
 pinMode(pin_temp_power, INPUT);   
} 


void read_temp(void)
{
 dht11 DHT11;

 
 int chk = DHT11.read(pin_temp_signal);
 
// Serial.print("Read sensor: ");
/*
 switch (chk)
 {
    case DHTLIB_OK: 
		Serial.println("OK"); 
		break;
    case DHTLIB_ERROR_CHECKSUM: 
		Serial.println("Checksum error"); 
		break;
    case DHTLIB_ERROR_TIMEOUT: 
		Serial.println("Time out error"); 
		break;
    default: 
		Serial.println("Unknown error"); 
		break;
 }
*/
// Serial.print("Humidity (%): ");
// Serial.println((float)DHT11.humidity, 2);

// Serial.print("Temperature (C): ");
// Serial.println((float)DHT11.temperature, 2);
 


// writeBCD(OregonMessageBuffer,8,2,DHT11.humidity);
 writeBCD(OregonMessageBuffer,10,2,DHT11.temperature);

 powerdown_temp();
}





/**********************************************************************************************/


volatile unsigned t1ovfcnt = 0;
volatile char oneshot=0;

#define BUFFER_SIZE 96

volatile unsigned long int buffer[BUFFER_SIZE];
volatile byte head = 0;
volatile byte tail = 0;

unsigned long basetime;
unsigned long echotime;
unsigned long phasetime;

byte adcsras;
byte adcsrbs;
byte admuxs;
byte acsrs;


void clearus(void)
{

 pinMode(TRIGGER_PIN, INPUT);
 pinMode(CHARGE_PIN, INPUT);
  
 ADCSRA=adcsras;
 ADCSRB=adcsrbs;
 ADMUX=admuxs;
 ACSR=acsrs;
}


 
void setupus(void) 
{
 adcsras=ADCSRA;
 adcsrbs=ADCSRB;
 admuxs=ADMUX;
 acsrs=ACSR;

 ADCSRA=0; // ADC Control and Status Register A
 ADCSRB=(1<<ACME); // enable multiplxer
 ADMUX=0; // ADC Multiplexer Selection Register
 ACSR=(1<<ACIC); // Analog Comparator Input Capture by Timer/Counter 1: On
  
 pinMode(TRIGGER_PIN, OUTPUT);
 pinMode(CHARGE_PIN, OUTPUT);
 pinMode(AIN0, INPUT);
 pinMode(AIN1, INPUT);
 DIDR1  |=  (1<< AIN0D);         // disable digital input buffer AIN0/1
 DIDR1  |=  (1<< AIN1D);

  
 TCCR1A = 0x0;    // put timer1 in normal mode
 TCCR1B = 0x2;    // change prescaler to divide clock by 8

 // clear any pending capture or overflow interrupts
 TIFR1 = (1<<ICF1) | (1<<TOV1);
 // Enable input capture and overflow interrupts
 TIMSK1 |= (1<<ICIE1) | (1<<TOIE1);



}


void read_sensor(void)
{
 setupus();
 powerup_sensor();
 delay(1000);
 byte newhead;
 unsigned long time1;
 int i;
 
 for(i=0;i<5;i++)
 {
  digitalWrite(CHARGE_PIN, 1); 
  delay(50);
  digitalWrite(CHARGE_PIN, 0); 
 
  time1=millis();
 
  digitalWrite(TRIGGER_PIN, 0);
  delayMicroseconds(4); 
  digitalWrite(TRIGGER_PIN, 1);   
  delayMicroseconds(10); 
  digitalWrite(TRIGGER_PIN, 0);   


  TCCR1B &= ~(1 << ICES1);   //switch edge detector to detect the fall next time
  ADMUX=0; // HC-SR04 echo pin to analogue comparator
  delayMicroseconds(10);

  head=tail;
  while(1)
  {
   if(head!=tail)
   {
    newhead = (head + 1) % BUFFER_SIZE;
    basetime=buffer[newhead];
    break;
   }
   if((millis()-time1)>50) break;
  }

  delayMicroseconds(5);
 
  TCCR1B |= (1<< ICES1);    //switch edge detector to detect the rise
  ADMUX=1;  // Envelope detector to analogue comparator
  delayMicroseconds(10);
 
  head=tail;
  while(1)
  {
   if(head!= tail) 
   {
//   digitalWrite(DEBUG_PIN, 1);
    newhead = (head + 1) % BUFFER_SIZE;
    echotime=buffer[newhead];
    break;
   }
   if((millis()-time1)>50) break;
  }

  delayMicroseconds(50);
  oneshot=1;
  ADMUX=2;  // Final amplifier output to analogue comparator
 
  head=tail;
  while(1)
  {
   if(head!= tail) 
   {
    newhead = (head + 1) % BUFFER_SIZE;
    phasetime=buffer[newhead];
    break;
   }
   if((millis()-time1)>50) break;
  }

  ADMUX=0; 
  oneshot=0;

// digitalWrite(DEBUG_PIN, 0);

  int phase=((phasetime-basetime)/(unsigned long)2) % (unsigned long)25;
  int distance=((unsigned long)750)-((echotime-basetime)/(unsigned long)2);

//  pr("phase %d dist %d %d",phase,distance,echotime-basetime);

  writeBCD(OregonMessageBuffer,12,4,distance);
  writeBCD(OregonMessageBuffer,8,2,phase);
 }
 
 clearus();
 powerdown_sensor();
}


ISR(TIMER1_OVF_vect) 
{ 
 t1ovfcnt++;              // keep track of overflows
}


ISR(TIMER1_CAPT_vect) 
{
 unsigned long t1temp;
 if(oneshot) ADMUX=0;
 tail=(tail+1) % BUFFER_SIZE;
 t1temp = ICR1;
 buffer[tail] = ((unsigned long)t1ovfcnt << 16) | t1temp;
}




void powerup_sensor(void)
{
 pinMode(pin_sensor_power, OUTPUT);  
 digitalWrite(pin_sensor_power, HIGH);
}  


void powerdown_sensor(void)
{
 pinMode(pin_sensor_power, INPUT);     
 digitalWrite(pin_sensor_power, LOW);
}  






/*******************************************************************************************/


// voltage 2 5.2
// temp    2 10.0
// hum     2 66
// freq    4
// 3,4,5,6,7 == 10 nybbles


void payload(void)
{
  
// Serial.print("millis ");  
// Serial.println(millis());


 long vcc=readVcc()/100;
 

 if(vcc<33)
 {
  pinMode(SWITCH_PIN, OUTPUT);
  digitalWrite(SWITCH_PIN, LOW);
  return;
 }

// Serial.print("Vcc "); 
// Serial.println(vcc,DEC);

 writeBCD(OregonMessageBuffer,6,2,vcc);
 
// Serial.print("Core temp ");
// Serial.println(readTemp(),DEC);
 
 read_sensor();

 read_temp();

 transmitter_send();

 if(!mode_slow)
 {
  pinMode(pin_led, OUTPUT);
  digitalWrite(pin_led, LOW);
  delay(100);
  digitalWrite(pin_led, HIGH);
  delay(100); 
  pinMode(pin_led, INPUT);  
  Serial.flush();
 }
 
}  



void power_down(void)
{
  

  
}

void power_up(void)
{
 powerup_temp();

 transmitter_setup();

// delay(1000); // temp sensor needs a second before it can be read
}


void setup() 
{
 Serial.begin(9600);        // connect to the serial port
}



void loop() 
{
 int i; 

 power_down();

 for(i=0;i<mode_speed;i++)
 {
  LowPower.powerDown(SLEEP_8S, ADC_OFF, BOD_OFF);
 }


// delay(150);

 power_up();

 payload();

 
 if(mode_count++==mode_switch)
 {
  mode_speed=mode_slowspeed;
  mode_slow=1;
 }
}


